'' *****************************
'' *  Wheel_Operations V1.0    *
'' *****************************

CON
 
 'I2C EEPROM Pin Assignments
 '--------------------------
  i2cSCL         = 28         'SCL Propeller P28 --- Same as bootup EEPROM on demo board.
' i2cSDA         = 29         'SDA is always assumed to be one pin# higher ! 
 
 'Motor Controller pin# assignments
 '--------------------------------- 
  Wheel_DirPin  = 0  '
  Wheel_StepPin = 1  '

  Spin_DirPin   = 2  '
  Spin_StepPin  = 3  '

 'MA3 Angle Encoder Input Pin Assignment
 '--------------------------------------
  pinAngPWM       = 24       'via Mouse PS/2 receptacle

 'For saving settings through a power cycle
 '-----------------------------------------
  EEPROM_Data_Size = 5  '# of items to be saved
  ee_zero_angle    = 2  'element # of EEPROM_Data[] assigned for storing zero angle

 'Operation States
 '----------------
  opStopped = 0
  opRunning = 1
  opPaused = 2
  opDwell = 3
 
VAR

  long  cogon, cog                                     ' cog running main sequence
  long  cogont, cogt                                   ' cog running timer

  long  TimerStack[200]                                ' timer stack    
  long  OpStack[200]                                   ' sequence stack

  long  running                                        ' operation status
  long  opState
  
  long  wheel_pos    'index position of sample wheel
  
 'This array holds data saved to EEPROM during a previous session 
  long  EEPROM_Data[EEPROM_Data_Size]

 'Motor Wheel Ramp Up/Down & Speed Setting  
  Long Speed_RadsPerSec 
  Long Accel_RadsSqSec

 'The longest time to wait for completion of a wheel (MW) movement before timing out.
  Long TimeOutSecs

  Long DwellTime[13]  'Holds dwell time at each wheel position in seconds
                      'array has elements 0 thru 12 

  long TimerInterval   'Clock cycles between timer tests in timer loop
  long TimerCount      'Number of counter intervals in dwell time
  long TimerSecs       'Seconds to wait -  dwell time interval
  long TimerValue      'Remaining timer counts in down count sequence ... in units of 1/100th of a second
  long TimerState      'FALSE(0)=PAUSED, TRUE(<>0)=RUNNING
   
 
OBJ

  MS           : "StepperMotorControl3"       '1 Cog - 'Spin' Stepper Driver
  
  MW           : "StepMotor_RampControl1"     '1 Cog - but only during stepper motion ...
                                              'Wheel' Stepper Driver with Accel/Decel/Speed Control
                                             
  AngleSensor  : "MA3P12_Encoder_V3"          '1 Cog - but only for ~10 mS at a time ....

  eeprom       : "24LC256_Microchip_EEPROM2"  'Cogless   

  serial       : "SerialMirror2"              'No additional cogs used, startx() method call is not required in this object 
 
  
PUB init : cog_ok | p

  opState := opStopped
  wheel_pos := -1   'unknown
  
 'Initialize EEPROM communication
 '-------------------------------
  eeprom.initialize(i2cSCL,0)        'Note: at startup EEPROM_Data[] array will hold previously saved data

 'Configure 'Spin' Stepper Motor Driver
 '-------------------------------------
  cog_ok := MS.Init(Spin_StepPin, Spin_DirPin)
  MS.SetPulsesPerRev(5500)    '2000 microsteps/rev * ~2.75 gear ratio = 5500 steps/spin-wheel revolution
  MS.SetSpeed_DegsPerSec(24)  '24 degs/sec = 1 rev/15 secs = 4 rpm
 
 'Configure 'Wheel' Stepper Motor Driver
 '-------------------------------------
  MW.Init(Wheel_StepPin, Wheel_DirPin)   'default timeout for movement is 20 seconds, use .SetTimeOutSecs method to alter

 'MW.Move(xStep,xAccel,xDecel,xSpeed)
  Accel_RadsSqSec  := 50    'default value
  Speed_RadsPerSec :=  100  'default value

 '@@@@@@@@@@@@@@@@@@@@@@@@@@
  TimeOutSecs := 10 'seconds
 '@@@@@@@@@@@@@@@@@@@@@@@@@@
  
  repeat p from 1 to 12 
     SetDwellTime(p,p)  '1 to 12 second dwell times for corresponding wheel positions
 
 'Initialize Angular Encoder
 '--------------------------
  AngleSensor.Start(pinAngPWM)
  AngleSensor.SetZeroAngle( EEPROM_Data[ee_zero_angle] )  'recover last saved zero-angle

    
PUB angle_read

    return AngleSensor.GetAngle         

PUB angle_set_zero

    AngleSensor.SetZeroAngle_ToCurrentAngle

    EEPROM_SaveData(ee_zero_angle, AngleSensor.GetZeroAngle)   'write zero angle to EEPROM now

    wheel_pos := 1   'define curent angle as position #1

    
PUB wheel_move_mdegs(mdegs) | nsteps  'move motor in units of millidegrees

   'With 8:1 gear ratio, 200 step/rev stepper, and 10 usteps/step motor drive:
   'Steps/Rev = 8 * 200 * 10 = 16000
   '== 400 steps per 9000 millidegrees, or 4 steps per 90 millidegrees

    nsteps :=  (mdegs * 4)/90

    serial.str(string("Calculated Motor Steps : "))
    serial.dec(nsteps)
    serial.putc(13)
   
    wheel_move_steps(nsteps)

PUB wheel_move_steps(wsteps)

    MW.Move(wsteps,Accel_RadsSqSec,Accel_RadsSqSec,Speed_RadsPerSec)    'default 20 second timeout on movement

PUB spin_speed (degs_per_sec)

    MS.SetSpeed_DegsPerSec(degs_per_sec)  'Default: 24 degs/sec = 1 rev/15 secs = 4 rpm
    

PUB spin_move_angle (mdegs)

    MS.SetDist_Angle(mdegs)     'set rotation angle (degs)
    MS.Move                     'enable move

PUB spin_dir_CW

    MS.SetDir_CW  

PUB spin_dir_CCW

    MS.SetDir_CCW  

PUB spin_go
      
    MS.SetDist_Pulses(0)        'force continuous, forever spin
    MS.Move

PUB spin_halt

    MS.halt

PUB timer_set (tsecs)

    TimerInterval := clkfreq/100    '10 mS
    TimerSecs := tsecs              'timer setting in seconds
    TimerCount := 100 * TimerSecs   '100 counts/sec * TimerSecs

PUB timer_secs_remaining

    return TimerValue/100   'return in units of seconds
  
PUB timer_start : okay

    TimerState := FALSE
    timer_stop   'only stops a cog executing our spin timer code
    
    okay := cogont := (cogt := cognew(TimerOperation, @TimerStack)) > 0

    if okay
       
       serial.str(string("Timer started. Running from Cog #"))
       serial.dec(cogt)
       serial.putc(13)

PUB timer_pause

    TimerState := FALSE

PUB timer_resume

    TimerState := TRUE

PUB timer_stop

'' Unload timer object - frees a cog
    if cogont~                       ' if object running, mark stopped
      cogstop(cogt)                  ' stop the cog

PUB TimerOperation | tstart

    TimerState := TRUE         'running
     
    TimerValue := TimerCount   'preload with wait time in units of a timer interval

    tstart := cnt
    repeat
      if TimerState == TRUE
         WaitCnt(tstart += TimerInterval)  '10 mS timer
         if TimerValue-- =< 0              'counting down ...
            TimerState := FALSE
            serial.str(string("Timer Expired Normally.",13))
            quit
       
PUB start : okay

   'Load wheel-sequence operation into a separate cog, if available

    stop    'running cog & halt spin stepper

    OutSignals_Enable(FALSE)  'float out signal lines prior to CogOperation taking control.

    okay := cogon := (cog := cognew(CogOperation, @OpStack)) > 0

    if okay
        
       running := TRUE 'begin operation execution immediately
       opState := opRunning
       
       serial.str(string("Wheel_Op started. Running from Cog #"))
       serial.dec(cog)
       serial.putc(13)
       
    else

       serial.str(string("Wheel_Op failed to start.",13))
       OutSignals_Enable(TRUE) 're-assert control over out signal lines
   

PUB stop

'' Unload timer object - frees a cog

    if cogon~                        ' if object running, mark stopped
      cogstop(cog)                   ' stop the cog

    opState := opStopped

    OutSignals_Enable(TRUE)   'Force output signal lines low while no operation is in progress.
                              'Otherwise they will be floating & their state indeterminate.
    
    MS.halt  'just in case spin stepper was left running at termination of CogOperation
 

PUB reset

'' Put timer on hold and clear registers

   running := TRUE
    
  ' pause                                                 ' stop timer
  ' tix := scs := mns := hrs := dys := 0                  ' reset registers


PUB pause

'' Put timer on hold without clearing registers

   running := FALSE  ' disable wheel motion sequence (set to false)


PUB run

'' Allow timer to run and update timer registers

   running := TRUE   ' enable wheel motion sequence (set to true)

 
PUB GetWheelPos

    return wheel_pos

PUB SetWheelPos(wpos)

    wheel_pos := wpos    'must only be called AFTER zero angle, position #1 has been defined !!!
    
    
PUB GetOpState

    return  opState


PUB SetDwellTime ( wh_pos, dwell_secs)

    DwellTime[wh_pos] := dwell_secs
    

PUB GetDwellTime ( wh_pos )

    return DwellTime[wh_pos]
    
              
PRI CogOperation | rcnt, cangle, tangle, mangle, i       

    OutSignals_Enable(TRUE)'Defines outputs pins to be used for Analog & Valid signals
                           'This method must be called from within CogOperation as it defines outputs for
                           'the spin interpreter running from THIS cog
 
    AnalogOutSignal(0)
    ValidOutSignal(0)

    if wheel_pos <> 1    'a CALIBRATION operation MUST proceed the start of sequence operation.
       OpState := OpStopped
       return
    
    rcnt := 0
       
    repeat

       rcnt += 1
                  
       if !running

         opState := opPaused
           
       else  'operation running ....

         opState := opRunning

         serial.str(string("======= Op Index # "))
         serial.dec(i)
         serial.putc(13)
   
        '==========================
        'MOVE TO NEW WHEEL POSITION
        '==========================
        
         repeat 
           WaitCnt(20_000_000 + cnt)       ' 0.25 seconds delay between angle sensor calls    
           cangle := AngleSensor.GetAngle  ' in millidegrees
         while (cangle <0) or (cangle > 360000)

         tangle := (wheel_pos - 1) *30000  ' multiply index# by 30 degrees (in units of millidegrees)

         mangle := tangle - cangle         ' calc move angle in millidegrees ( target_angle - current_angle)
         if mangle < 0
            mangle += 360000               ' correct for angle rollover

         wheel_move_mdegs(mangle)  

        '=================
        'ENGAGE SPIN MOTOR  
        '=================
         
         WaitCnt(clkfreq + cnt)        'pause prior to spin wheel engagement
         spin_go                       'START SPIN MOTOR
         AnalogOutSignal(wheel_pos)
         ValidOutSignal(1)

        '============================================================
        'Dwell Time of Spin Operation  'WAIT FOR SPECIFIED DWELL TIME
        '============================================================
         'WaitCnt(clkfreq*3 + cnt)
         
         timer_set(DwellTime[wheel_pos]) 'seconds of dwell time
         timer_start

         opState := OpDwell
         repeat while TimerState==TRUE
            'test for paused state
            
        '===============
        'HALT SPIN MOTOR 
        '===============          
         spin_halt                     'STOP SPIN MOTOR
         ValidOutSignal(0)
         AnalogOutSignal(0)

        '=================================================================
        'ADVANCE TO NEXT WHEEL POSITION (WITH NON-ZERO DWELL TIME SETTING)
        '=================================================================

         repeat
           if ++wheel_pos > 12  
             wheel_pos := 1
         until DwellTime[wheel_pos] <> 0

PRI OutSignals_Enable(out_state)

 If out_state == TRUE

   'Analog Out Signal
   '=================
    Dira[19..16] := %1111 'define as output pins
    Outa[19..16] := %0000 'initialize at zero

   'Valid Logic Signal
   '==================
    Dira[7] := %1 'define as output pins
    Outa[7] := %0 'initialize at zero

 Else   'FALSE

    Dira[19..16] := %0000 'define as floating input pins
    Dira[7] := %0         'define as a floating input pin
  

PUB AnalogOutSignal(i) | j    'output an analog ~1v max signal via the VGA monitor pins of the demo board

  'Dira[19..16] := %1111 'define as output pins
  'Outa[19..16] := %0000 'initialize at zero

  i //= 15      'restrict input range to legal values between 0 and 14. Value 15 illegal because it will
                'be incremented by one in the code below to 16 which is too high.
  
  if i > 7
     j := i+1   'necessary to "bump up" the voltage by one step in order to achieve good voltage level spacings.
  else
     j := i   

  Outa[19..16] := j    'resistive summing junction of 4 digital outputs creates a composite analog voltage.
           
PUB ValidOutSignal(i) | j    'output an analog ~1v max signal via the VGA monitor pins of the demo board

  'Dira[7] := %1 'define as output pins
  'Outa[7] := %0 'initialize at zero
    
  serial.str(string("Valid Output Signal State: "))
  serial.dec(i)
  serial.putc(13)
      
  Outa[7] := i  'TTL compatible logic output level
                 
PUB EEPROM_Save (array_index)

 ' Since EEPROM addresses directly correspond to the addresses of
 ' Propeller RAM, a .Write_Block call with identical eeAddress and
 ' ramAddress parameters will only overwrite the data range specified,
 ' not firmware boot code elsewhere.

 ' eeAddress  == EEPROM address
 ' ramAddress == Propeller internal RAM address 

 'READ EXAMPLE:
 ' Read from EEPROM to RAM a block of LONGs:
 ' Read_Block(eeAddress, ramAddress, ByteCount)
 '----------------------------------------------
 ' eeprom.Read_Block(@EEPROM_Data, @EEPROM_Data, EEPROM_Data_Size*4) 
   
 'WRITE EXAMPLE: 
 ' Write to EEPROM from RAM a block of LONGs:
 ' Write_Block(eeAddress, ramAddress, ByteCount)
 '----------------------------------------------  
 ' eeprom.Write_Block(@EEPROM_Data, @EEPROM_Data, EEPROM_Data_Size*4)  

 'Save 1 LONG (4 bytes) defined by 'array_index' to EEPROM 
  eeprom.Write_Block(@EEPROM_Data[array_index], @EEPROM_Data[array_index], 4)

PUB EEPROM_SaveAll

  eeprom.Write_Block(@EEPROM_Data, @EEPROM_Data, EEPROM_Data_Size*4)  

PUB EEPROM_SaveData(array_index, data_value)

 'Write data to array
  EEPROM_Data[array_index] := data_value
    
 'Save 1 LONG (4 bytes) defined by 'array_index' to EEPROM 
  eeprom.Write_Block(@EEPROM_Data[array_index], @EEPROM_Data[array_index], 4)
 
  